#include "precompile.h"

#include <string.h>

#include "ss_proto.pb.h"
#include "ss_msgid.pb.h"
#include "mastersvr.h"
#include <a8/tcpclient.h>
#include <a8/asynctcpclient.h>
#include <a8/ioloop.h>
#include <a8/udplog.h>
#include <a8/timer.h>
#include "app.h"

const int PACK_MAX = 1024 * 64;

void MasterSvr::Init(int instance_id, const std::string& remote_ip, int remote_port)
{
    this->instance_id = instance_id;
    this->remote_ip = remote_ip;
    this->remote_port = remote_port;

    recv_bufflen_ = 0;
    last_pong_tick = a8::XGetTickCount();
    recv_buff_ = (char*) malloc(PACK_MAX * 2);
#if ASYNC_TCPCLIENT
    tcp_client_ = a8::IoLoop::Instance()->CreateAsyncTcpClient();
#else
    tcp_client_ = new a8::TcpClient();
#endif
    tcp_client_->remote_address = remote_ip;
    tcp_client_->remote_port = remote_port;
    tcp_client_->on_error = std::bind(&MasterSvr::on_error, this, std::placeholders::_1, std::placeholders::_2);
    tcp_client_->on_connect = std::bind(&MasterSvr::on_connect, this, std::placeholders::_1);
    tcp_client_->on_disconnect = std::bind(&MasterSvr::on_disconnect, this, std::placeholders::_1);
    tcp_client_->on_socketread = std::bind(&MasterSvr::on_socketread, this ,std::placeholders::_1, std::placeholders::_2, std::placeholders::_3);
    timer_ = a8::Timer::Instance()->AddRepeatTimer(1000 * 9 + a8::RandEx(500, 150),
                                                  a8::XParams().SetSender(this),
                                                  [] (const a8::XParams& param)
                                                  {
                                                      MasterSvr* conn = (MasterSvr*)param.sender.GetUserData();
                                                      conn->CheckAlive();
                                                  });
}

void MasterSvr::UnInit()
{
    a8::Timer::Instance()->DeleteTimer(timer_);
    timer_ = nullptr;
    tcp_client_->Close();
#if ASYNC_TCPCLIENT
    a8::IoLoop::Instance()->DestoryAsyncTcpClient(tcp_client_);
#else
    delete tcp_client_;
#endif
    tcp_client_ = nullptr;
    recv_bufflen_ = 0;
    free(recv_buff_);
    recv_buff_ = nullptr;
}

void MasterSvr::Open()
{
    tcp_client_->Open();
}

void MasterSvr::Close()
{
    tcp_client_->Close();
}

bool MasterSvr::Connected()
{
    return tcp_client_->Connected();
}

#if ASYNC_TCPCLIENT
void MasterSvr::on_error(a8::AsyncTcpClient* sender, int errorId)
#else
void MasterSvr::on_error(a8::TcpClient* sender, int errorId)
#endif
{
    a8::UdpLog::Instance()->Error("MasterSvr errorid=%d remote_ip:%s remote_port:%d",
                                  {
                                   errorId,
                                   sender->remote_address,
                                   sender->remote_port
                                  });
}

#if ASYNC_TCPCLIENT
void MasterSvr::on_connect(a8::AsyncTcpClient* sender)
#else
void MasterSvr::on_connect(a8::TcpClient* sender)
#endif
{
    recv_bufflen_ = 0;
    a8::UdpLog::Instance()->Info("masterserver connected", {});
}

#if ASYNC_TCPCLIENT
void MasterSvr::on_disconnect(a8::AsyncTcpClient* sender)
#else
void MasterSvr::on_disconnect(a8::TcpClient* sender)
#endif
{
    recv_bufflen_ = 0;
    a8::UdpLog::Instance()->Info("masterserver %d disconnected after 10s later reconnect", {instance_id});
    App::Instance()->AddIMMsg(IM_MasterSvrDisconnect,
                              a8::XParams()
                              .SetSender(instance_id)
                              );
}

#if ASYNC_TCPCLIENT
void MasterSvr::on_socketread(a8::AsyncTcpClient* sender, char* buf, unsigned int len)
#else
void MasterSvr::on_socketread(a8::TcpClient* sender, char* buf, unsigned int len)
#endif
{
    #if 0
    ++App::Instance()->perf.read_count;
    #endif
    if (recv_bufflen_ + len > 2 * PACK_MAX) {
        recv_bufflen_ = 0;
        a8::UdpLog::Instance()->Debug("recvied masterserver too long message", {});
        return;
    } else {
        memmove(&recv_buff_[recv_bufflen_], buf, len);
        recv_bufflen_ += len;
    }

    bool warning = false;
    unsigned int offset = 0;
    while (recv_bufflen_ - offset >= sizeof(f8::PackHead)) {
        f8::PackHead* p = (f8::PackHead*) &recv_buff_[offset];
        if (p->magic_code == f8::MAGIC_CODE) {
            if (recv_bufflen_ - offset < sizeof(f8::PackHead) + p->packlen) {
                break;
            }
            App::Instance()->AddSocketMsg(SF_MasterServer,
                                          0,
                                          instance_id,
                                          p->msgid,
                                          p->seqid,
                                          &recv_buff_[offset + sizeof(f8::PackHead)],
                                          p->packlen);
            offset += sizeof(f8::PackHead) + p->packlen;
        } else {
            warning = true;
            offset++;
            continue;
        }
    }

    if (warning) {
        a8::UdpLog::Instance()->Debug("recvied bad package", {});
    }
    if (offset > 0 && offset < recv_bufflen_) {
        memmove(recv_buff_, recv_buff_ + offset, recv_bufflen_ - offset);
    }
    recv_bufflen_ -= offset;
    #if 1
    last_pong_tick = a8::XGetTickCount();
    #endif
}

void MasterSvr::CheckAlive()
{
    if (!Connected()) {
        Open();
    } else {
        if (a8::XGetTickCount() - last_pong_tick > 60 * 10 * 1000) {
            last_pong_tick = a8::XGetTickCount();
            Close();
            Open();
        } else {
            ss::SS_Ping msg;
            SendMsg(msg);
        }
    }
}
